package ic.opt;

import ic.ast.*;

/**
 * To create an optimization, simply subclass this class and define
 * optimize(md) to compute any dataflow information necessary to
 * perform the optimization on md and then modify the method's TAC
 * list. You can either provide methods in your TACList class to
 * modify an existing list, or you can construct a new list to replace
 * the old one.
 * 
 * The methods return true to indicate that the optimization changed
 * the TAC in some way. While not required, you can provide a -iter
 * command-line option that iterative applies all optimizations over
 * and over again until no additional changes happen.
 */
public abstract class Optimization {

	/**
	 * Apply the optimization to each Method in p.
	 * Return true if the TAC for any of them changed.
	 */
	public boolean optimize(Program p) {
		boolean changed = false;

		/* 
		 * Insert your own code to call optimize on each MethodDecl.
		 * You can use a visitor, or just iterate over the decls in each
		 * ClassDecl and look for MethodDecls.
		 */

		for (ClassDecl c : p.getClasses()) {
			for (TypeDeclElem elem : c.getElems()) {
				if (elem instanceof MethodDecl) {
					if (optimize((MethodDecl) elem)) {
						changed = true;
					}
				}
			}
		}

		return changed;
	}


	/**
	 * Apply the optimization to the method md.
	 * Return true if the TAC for the method changes
	 * at all.
	 */
	public abstract boolean optimize(MethodDecl md);
}
